"""Functions related to vectors and field expansion in the FMM scheme.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import dataclasses
import enum
from typing import Any, Tuple

import jax.numpy as jnp
import numpy as onp
from jax import tree_util

from fmmax import utils

NDArray = onp.ndarray[Any, Any]

# Officially defines the x- and y- directions. By convention, the x-axis preceeds
# the y-axis in our array indexing scheme.
X: jnp.ndarray = jnp.array([1.0, 0.0])
Y: jnp.ndarray = jnp.array([0.0, 1.0])


@dataclasses.dataclass
class LatticeVectors:
    """Stores a pair of lattice vectors.

    Note that this is just a pair of 2-dimensional vectors, which may be either for
    the real-space lattice or the reciprocal space lattice, depending on usage.

    Attributes:
        u: The first primitive lattice vector.
        v: The second primitive lattice vector, with identical shape.
    """

    u: jnp.ndarray
    v: jnp.ndarray

    def __post_init__(self) -> None:
        if isinstance(self.u, jnp.ndarray):
            if self.u.shape[-1] != 2 or self.v.shape[-1] != 2:
                raise ValueError(
                    f"`u` and `v` must have a trailing length of 2, but got shapes "
                    f"{self.u.shape} and {self.v.shape}."
                )

    @property
    def reciprocal(self) -> "LatticeVectors":
        """Returns the corresponding vectors on the reciprocal lattice."""
        return reciprocal(self)


@dataclasses.dataclass
class Expansion:
    """Stores the expansion.

    The expansion consists of the integer coefficients of the reciprocal lattice vectors
    used in the Fourier expansion of fields, permittivities, etc. in the FMM scheme.

    Attributes:
        basis_coefficients: The integer coefficients of the primitive reciprocal lattice
            vectors, which generate the full set of reciprocal-space vectors in the
            expansion.
        num_terms: The number of terms in the expansion.
    """

    basis_coefficients: NDArray

    def __post_init__(self) -> None:
        if self.basis_coefficients.ndim != 2 or self.basis_coefficients.shape[-1] != 2:
            raise ValueError(
                f"`basis_coefficients` must have shape `(num, 2)` but got "
                f"{self.basis_coefficients.shape}."
            )

    @property
    def num_terms(self) -> int:
        return self.basis_coefficients.shape[-2]


@enum.unique
class Truncation(enum.Enum):
    """Enumerates truncation modes."""

    CIRCULAR = "circular"
    PARALLELOGRAMIC = "parallelogramic"


def generate_expansion(
    primitive_lattice_vectors: LatticeVectors,
    approximate_num_terms: int,
    truncation: Truncation,
) -> Expansion:
    """Generates the expansion for the specified real-space basis.

    Args:
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        approximate_num_terms: The approximate number of terms in the expansion. To
            maintain a symmetric expansion, the total number of terms may differ from
            this value.
        truncation: The truncation to be used for the expansion.

    Returns:
        The `Expansion`. The basis coefficients of the expansion are sorted so that
        the zeroth-order term is first.
    """
    reciprocal_vectors = primitive_lattice_vectors.reciprocal
    if truncation == Truncation.CIRCULAR:
        basis_coefficients = _basis_coefficients_circular(
            reciprocal_vectors, approximate_num_terms
        )
    elif truncation == Truncation.PARALLELOGRAMIC:
        basis_coefficients = _basis_coefficients_parallelogramic(
            reciprocal_vectors, approximate_num_terms
        )
    else:
        raise ValueError(f"Unknown `truncation`, got {truncation}.")
    return Expansion(basis_coefficients)


def reciprocal(lattice_vectors: LatticeVectors) -> LatticeVectors:
    """Computes the reciprocal vectors for the `basis`."""
    cross_product = _cross_product(lattice_vectors.u, lattice_vectors.v)
    uprime = (
        jnp.stack([lattice_vectors.v[..., 1], -lattice_vectors.v[..., 0]], axis=-1)
        / cross_product
    )
    vprime = (
        jnp.stack([-lattice_vectors.u[..., 1], lattice_vectors.u[..., 0]], axis=-1)
        / cross_product
    )
    return LatticeVectors(u=uprime, v=vprime)


def unit_cell_coordinates(
    primitive_lattice_vectors: LatticeVectors,
    shape: Tuple[int, int],
    num_unit_cells: Tuple[int, int],
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute spatial coordinates given the grid shape and number of unit cells."""
    i_stop = num_unit_cells[0] * shape[0]
    j_stop = num_unit_cells[1] * shape[1]
    i, j = tuple(
        jnp.meshgrid(
            jnp.arange(0, i_stop) / shape[0],
            jnp.arange(0, j_stop) / shape[1],
            indexing="ij",
        )
    )
    x = (
        i * primitive_lattice_vectors.u[..., jnp.newaxis, jnp.newaxis, 0]
        + j * primitive_lattice_vectors.v[..., jnp.newaxis, jnp.newaxis, 0]
    )
    y = (
        i * primitive_lattice_vectors.u[..., jnp.newaxis, jnp.newaxis, 1]
        + j * primitive_lattice_vectors.v[..., jnp.newaxis, jnp.newaxis, 1]
    )
    return x, y


# -----------------------------------------------------------------------------
# Functions related to transverse wavevectors.
# -----------------------------------------------------------------------------


def plane_wave_in_plane_wavevector(
    wavelength: jnp.ndarray,
    polar_angle: jnp.ndarray,
    azimuthal_angle: jnp.ndarray,
    permittivity: jnp.ndarray,
) -> jnp.ndarray:
    """Computes the in-plane wavevector for a plane-wave excitation.

    Args:
        wavelength: The free-space wavelength of the plane-wave excitation.
        polar_angle: Polar angle of the plane-wave excitation.
        azimuthal_angle: Azimuthal angle of plane-wave the excitation.
        permittivity: Scalar permittivity of the medium in which the polar angle
            and azimuthal angle are specified.

    Returns:
        The fundamental transverse wavevector, i.e. `(kx0, ky0)`.
    """
    angular_frequency = utils.angular_frequency_for_wavelength(wavelength)
    kx0 = (
        angular_frequency
        * jnp.sin(polar_angle)
        * jnp.cos(azimuthal_angle)
        * jnp.sqrt(permittivity.real)
    )
    ky0 = (
        angular_frequency
        * jnp.sin(polar_angle)
        * jnp.sin(azimuthal_angle)
        * jnp.sqrt(permittivity.real)
    )
    return jnp.stack([kx0, ky0], axis=-1)


def brillouin_zone_in_plane_wavevector(
    brillouin_grid_shape: Tuple[int, int],
    primitive_lattice_vectors: LatticeVectors,
) -> jnp.ndarray:
    """Computes in-plane wavevectors on a regular grid in the first Brillouin zone.

    The wavevectors are found by dividing the Brillouin zone into a grid with the
    specified shape; the wavevectors are at the centers of the grid voxels.

    Args:
        brillouin_grid_shape: The shape of the wavevector grid.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.

    Returns:
        The in-plane wavevectors, with shape `brillouin_grid_shape + (2,)`.
    """
    if len(brillouin_grid_shape) != 2 or brillouin_grid_shape < (1, 1):
        raise ValueError(
            f"`brillouin_grid_shape` must be length-2 with positive values, "
            f"but got {brillouin_grid_shape}."
        )

    spacing_i = 1 / brillouin_grid_shape[0]
    spacing_j = 1 / brillouin_grid_shape[1]
    i, j = jnp.meshgrid(
        jnp.arange(-0.5 + spacing_i / 2, 0.5, spacing_i),
        jnp.arange(-0.5 + spacing_j / 2, 0.5, spacing_j),
        indexing="ij",
    )
    reciprocal_vectors = primitive_lattice_vectors.reciprocal
    return jnp.stack(
        [
            2
            * jnp.pi
            * (i * reciprocal_vectors.u[..., 0] + j * reciprocal_vectors.v[..., 0]),
            2
            * jnp.pi
            * (i * reciprocal_vectors.u[..., 1] + j * reciprocal_vectors.v[..., 1]),
        ],
        axis=-1,
    )


def transverse_wavevectors(
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: LatticeVectors,
    expansion: Expansion,
) -> jnp.ndarray:
    """Obtains all the relevant transverse wavevectors given the expansion.

    Args:
        in_plane_wavevector: The zeroth-order transverse wavevector.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        expansion: The expansion for which the set of transverse wavevectors are sought.

    Returns:
        The transverse wavevectors.
    """
    reciprocal_vectors = primitive_lattice_vectors.reciprocal
    kx = in_plane_wavevector[..., 0, jnp.newaxis] + 2 * jnp.pi * (
        expansion.basis_coefficients[:, 0] * reciprocal_vectors.u[..., 0]
        + expansion.basis_coefficients[:, 1] * reciprocal_vectors.v[..., 0]
    )
    ky = in_plane_wavevector[..., 1, jnp.newaxis] + 2 * jnp.pi * (
        expansion.basis_coefficients[:, 0] * reciprocal_vectors.u[..., 1]
        + expansion.basis_coefficients[:, 1] * reciprocal_vectors.v[..., 1]
    )
    return jnp.stack([kx, ky], axis=-1)


# -----------------------------------------------------------------------------
# Functions to generate basis coefficients with various truncations.
# -----------------------------------------------------------------------------


def _basis_coefficients_circular(
    primitive_lattice_vectors: LatticeVectors,
    approximate_num_terms: int,
) -> onp.ndarray:
    """Computes the basis coefficients of lattice vectors lying within a circle.

    The coefficients generate a set of lattice vectors from a basis using a circular
    truncation, so that all lattice vectors lying within a circular region with area
    given by `num * u ⨉ v` are included, where `⨉` is the cross product. The size of
    the set will generally be close to `approximate_num_terms`.

    Args:
        primitive_lattice_vectors: Primitive vectors for the reciprocal-space lattice.
        approximate_num_terms: The approximate number of terms in the expansion. To
            maintain a symmetric expansion, the total number of terms may differ from
            this value.

    Returns:
        The coefficients, with shape `(num_vectors, 2)`. The final axis gives the
        coefficient for the first and second vector in the basis.
    """
    # Generate candidate coefficients. These will include more coefficients than needed;
    # subsequently we will filter based on magnitude.
    g = onp.arange(-approximate_num_terms // 2, approximate_num_terms // 2 + 1)
    G1, G2 = onp.meshgrid(g, g, indexing="ij")
    G1 = G1.flatten()
    G2 = G2.flatten()

    # Generate the actual vectors and compute their magnitude.
    vectors = (
        G1[..., onp.newaxis] * primitive_lattice_vectors.u[..., onp.newaxis, :]
        + G2[..., onp.newaxis] * primitive_lattice_vectors.v[..., onp.newaxis, :]
    )
    magnitude = onp.linalg.norm(vectors, axis=-1)

    # Include all vectors lying within the circle with area equal to cross product
    # of `u` and `v` scaled by `num_terms`.
    max_magnitude = onp.sqrt(
        approximate_num_terms
        * onp.abs(
            _cross_product(primitive_lattice_vectors.u, primitive_lattice_vectors.v)
        )
        / onp.pi
    )
    mask = magnitude < max_magnitude

    G1 = G1[mask]
    G2 = G2[mask]
    magnitude = magnitude[mask]
    G = onp.stack([G1, G2], axis=-1)

    order = onp.argsort(magnitude, kind="stable")
    return G[..., order, :]


def _basis_coefficients_parallelogramic(
    primitive_lattice_vectors: LatticeVectors,
    approximate_num_terms: int,
) -> onp.ndarray:
    """Computes the basis coefficients of lattice vectors lying within a parallelogram.

    The coefficients generate a set of lattice vectors from a basis using a
    parallelogramic truncation. The size of the set will generally be close to
    `approximate_num_terms`.

    Args:
        primitive_lattice_vectors: Primitive vectors for the reciprocal-space lattice.
        approximate_num_terms: The approximate number of terms in the expansion. To
            maintain a full parallelogram expansion, the total number of terms may
            differ from this value.

    Returns:
        The coefficients, with shape `(num_vectors, 2)`. The final axis gives the
        coefficient for the first and second vector in the basis.
    """

    ku_spacing = onp.sqrt(onp.sum(onp.abs(primitive_lattice_vectors.u) ** 2))
    kv_spacing = onp.sqrt(onp.sum(onp.abs(primitive_lattice_vectors.v) ** 2))

    # Solve for `(nu, nv)` such that we approximately satisfy
    #     (2 * nu + 1) * (2 * nv + 1) = approximate_num_terms
    # and
    #     nu / nv = kv_spacing / ku_spacing
    # Since `ku_spacing` and `kv_spacing` give the spacing of points in reciprocal
    # space along the ku and kv directions, respectively, this second equation ensures
    # that we are chosing points in a parallelogram with equal-length sides in k-space.
    # (Note that while the sides of the parallelogram have equal length, the number of
    # points along each direction will differ, as these are spaced by `ku_spacing` and
    # `kv_spacing`.)

    def _solve_quadratic(ratio):
        a = 4 * ratio
        b = 2 * (ratio + 1)
        c = 1 - approximate_num_terms
        nu = (-b + onp.sqrt(b**2 - 4 * a * c)) / (2 * a)
        return int(onp.around(nu))

    nu = _solve_quadratic(ku_spacing / kv_spacing)
    nv = _solve_quadratic(kv_spacing / ku_spacing)

    G1, G2 = onp.meshgrid(
        onp.arange(-nu, nu + 1),
        onp.arange(-nv, nv + 1),
        indexing="ij",
    )
    G1 = G1.flatten()
    G2 = G2.flatten()
    G = onp.stack([G1, G2], axis=-1)
    # Generate the actual vectors and compute their magnitude.
    vectors = (
        G1[..., onp.newaxis] * primitive_lattice_vectors.u[..., onp.newaxis, :]
        + G2[..., onp.newaxis] * primitive_lattice_vectors.v[..., onp.newaxis, :]
    )
    magnitude = onp.linalg.norm(vectors, axis=-1)

    order = onp.argsort(magnitude, kind="stable")
    return G[..., order, :]


def _cross_product(x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Computes the cross product of 2D vectors `x` and `y`."""
    return x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]


# -----------------------------------------------------------------------------
# Register custom objects in this module with jax to enable `jit`.
# -----------------------------------------------------------------------------


tree_util.register_pytree_node(
    LatticeVectors,
    lambda lv: ((lv.u, lv.v), None),
    lambda _, uv: LatticeVectors(*uv),
)


tree_util.register_pytree_node(
    Truncation,
    lambda x: ((), x.value),
    lambda value, _: Truncation(value),
)


def unflatten_expansion(
    aux_tuple: Tuple["_HashableArray"],
    leaves: Tuple,
) -> Expansion:
    """Unflattens the `Expansion`."""
    del leaves
    wrapped: "_HashableArray" = aux_tuple[0]
    coeffs: onp.ndarray = wrapped.value
    return Expansion(basis_coefficients=coeffs)


tree_util.register_pytree_node(
    Expansion,
    lambda e: ((), (_HashableArray(e.basis_coefficients),)),
    unflatten_func=unflatten_expansion,
)


class _HashableArray:
    """Hashable wrapper for numpy arrays."""

    def __init__(self, value: onp.ndarray):
        self.value: onp.ndarray = value

    def __hash__(self):
        return hash((self.value.shape, self.value.dtype, self.value.tobytes()))

    def __eq__(self, other):
        if not isinstance(other, _HashableArray):
            return False
        return (
            self.value.shape == other.value.shape
            and self.value.dtype == other.value.dtype
            and onp.all(self.value == other.value)
        )



"""Functions related to fields in the FMM algorithm.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import functools
from typing import Callable, Sequence, Tuple

import jax.numpy as jnp

from fmmax import basis, fft, fmm, scattering, utils


def propagate_amplitude(
    amplitude: jnp.ndarray,
    distance: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
) -> jnp.ndarray:
    """Propagates waves with the given `amplitude` by `distance`.

    The propagation is along the wave direction, i.e. when `distance` is positive,
    amplitudes for forward-propagating waves are those associated with a positive
    shift along the z-axis, while the reverse is true for backward-propagating wave
    amplitudes.

    Args:
        amplitude: The amplitudes to be propagated, with a trailing batch dimension.
        distance: The distance to be propagated.
        layer_solve_result: The result of the layer eigensolve.

    Returns:
        The propagated amplitudes.
    """
    _validate_amplitudes_shape(
        (amplitude,),
        num_terms=2 * layer_solve_result.expansion.num_terms,
    )
    q = layer_solve_result.eigenvalues
    fd = jnp.exp(1j * q[..., jnp.newaxis] * distance)
    return amplitude * fd


def colocate_amplitudes(
    forward_amplitude_start: jnp.ndarray,
    backward_amplitude_end: jnp.ndarray,
    z_offset: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
    layer_thickness: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute the forward- and backward-propagating wave amplitudes at `z_offset`.

    The calculation is for a batch of amplitudes, with the batch dimension being
    final dimension.

    Args:
        forward_amplitude_start: The amplitude of the forward eigenmodes at the
            start of the layer, with a trailing batch dimension.
        backward_amplitude_end: The amplitude of the backward eigenmodes at the
            end of the layer.
        z_offset: The location where the colocated amplitudes are sought, as an
            offset from the start of the layer.
        layer_solve_result: The result of the layer eigensolve.
        layer_thickness: The thickness of the layer.

    Returns:
        The forward- and backward-propagating wave amplitudes at `z`.
    """
    _validate_amplitudes_shape(
        (forward_amplitude_start, backward_amplitude_end),
        num_terms=2 * layer_solve_result.expansion.num_terms,
    )
    return (
        propagate_amplitude(
            amplitude=forward_amplitude_start,
            distance=z_offset,
            layer_solve_result=layer_solve_result,
        ),
        propagate_amplitude(
            amplitude=backward_amplitude_end,
            distance=layer_thickness - z_offset,
            layer_solve_result=layer_solve_result,
        ),
    )


def _validate_amplitudes_shape(
    amplitudes: Tuple[jnp.ndarray, ...], num_terms: int
) -> None:
    """Validates the shapes for the `amplitudes`."""
    if (
        amplitudes[0].ndim < 2
        or amplitudes[0].shape[-2] != num_terms
        or not utils.batch_compatible_shapes(*[a.shape for a in amplitudes])
    ):
        raise ValueError(
            f"All amplitudes must have matching shape `(..., {num_terms}, "
            f"amplitudes_batch_size)` but got shapes {[a.shape for a in amplitudes]}."
        )


def amplitude_poynting_flux(
    forward_amplitude: jnp.ndarray,
    backward_amplitude: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Returns total Poynting flux for forward and backward eigenmodes.

    This function decomposes the total field into components associated with
    the forward and backward amplitudes, and returns the time-average flux in
    each order for these two components. The calculation follows section 5.1
    of [2012 Liu].

    In the general case, a forward eigenmode may actually have negative
    Poynting flux, and therefore the quantities computed by this function
    should not be interpreted as the total forward and backward flux, but only
    the total flux associated with the forward and backward eigenmodes.

    If the total forward and backward flux is desired, `directional_poynting_flux`
    should be used instead. This function should only be used in the specific
    case where the flux associated with the forward and backward eigenmodes is
    needed.

    Args:
        forward_amplitude: The amplitude of the forward eigenmodes, with a
            trailing batch dimension.
        backward_amplitude: The amplitude of the backward eigenmodes, at the
            same location in space as the `forward_amplitude`.
        layer_solve_result: The results of the layer eigensolve.

    Returns:
        The Poynting flux associated with the forward and backward eigenmodes.
    """
    _validate_amplitudes_shape(
        (forward_amplitude, backward_amplitude),
        num_terms=2 * layer_solve_result.expansion.num_terms,
    )

    alpha_h = layer_solve_result.eigenvectors @ forward_amplitude
    beta_h = layer_solve_result.eigenvectors @ backward_amplitude

    A = _poynting_flux_a_matrix(layer_solve_result)
    alpha_e = A @ forward_amplitude
    beta_e = A @ backward_amplitude

    s_forward = jnp.asarray(0.5, dtype=forward_amplitude.dtype) * (
        (jnp.conj(alpha_e) * alpha_h + jnp.conj(alpha_h) * alpha_e)
        + (jnp.conj(beta_h) * alpha_e - jnp.conj(beta_e) * alpha_h)
    )
    s_backward = jnp.asarray(0.5, dtype=forward_amplitude.dtype) * (
        -(jnp.conj(beta_e) * beta_h + jnp.conj(beta_h) * beta_e)
        + jnp.conj(jnp.conj(beta_h) * alpha_e - jnp.conj(beta_e) * alpha_h)
    )
    return jnp.real(s_forward), jnp.real(s_backward)


def directional_poynting_flux(
    forward_amplitude: jnp.ndarray,
    backward_amplitude: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Returns total forward and backward Poynting flux.

    This function decomposes the total field into components resulting from the
    the eigenmodes with positive and negative Poynting flux, and returns the
    time-average flux in each order for these two components. The calculation
    follows section 5.1 of [2012 Liu].

    In the general case, a forward eigenmode may actually have negative
    Poynting flux, and so e.g. it may occur that a one-hot forward amplitude
    vector yields zero forward flux and nonzero backward flux.

    If the flux associated with the forward and backward eigenmodes is desired,
    `amplitude_poynting_flux` should be used instead. This function serves the
    more typical case where the total forward flux and total backward flux is
    desired.

    Args:
        forward_amplitude: The amplitude of the forward eigenmodes, with a
            trailing batch dimension.
        backward_amplitude: The amplitude of the backward eigenmodes, at the
            same location in space as the `forward_amplitude`.
        layer_solve_result: The results of the layer eigensolve.

    Returns:
        The Poynting flux associated with the forward and backward eigenmodes.
    """
    eigenmode_flux = eigenmode_poynting_flux(layer_solve_result)
    fwd_flux_eigenmode = eigenmode_flux > 0
    fwd_flux_eigenmode = fwd_flux_eigenmode[..., jnp.newaxis]

    # The eigenmodes may not all have positive Poynting flux, i.e. there may be
    # eigenmodes which carry energy backward assoicated with `forward_amplitude`.
    # We break apart `forward_amplitude` and `backward_amplitude` to separate
    # those which carry energy forward from those which carry energy backward.
    forward_amplitude_fwd_eigenmode = jnp.where(
        fwd_flux_eigenmode, forward_amplitude, 0.0
    )
    forward_amplitude_bwd_eigenmode = jnp.where(
        fwd_flux_eigenmode, 0.0, forward_amplitude
    )
    backward_amplitude_fwd_eigenmode = jnp.where(
        ~fwd_flux_eigenmode, backward_amplitude, 0.0
    )
    backward_amplitude_bwd_eigenmode = jnp.where(
        ~fwd_flux_eigenmode, 0.0, backward_amplitude
    )

    A = _poynting_flux_a_matrix(layer_solve_result)
    phi = layer_solve_result.eigenvectors

    def _poynting_flux_subset(
        a_total: jnp.ndarray,
        b_total: jnp.ndarray,
        a_subset: jnp.ndarray,
        b_subset: jnp.ndarray,
    ) -> jnp.ndarray:
        return jnp.asarray(0.5) * (
            jnp.conj(A @ (a_total - b_total)) * (phi @ (a_subset + b_subset))
            + jnp.conj(phi @ (a_total + b_total)) * (A @ (a_subset - b_subset))
        )

    s_forward = _poynting_flux_subset(
        a_total=forward_amplitude,
        b_total=backward_amplitude,
        a_subset=forward_amplitude_fwd_eigenmode,
        b_subset=backward_amplitude_fwd_eigenmode,
    )
    s_backward = _poynting_flux_subset(
        a_total=forward_amplitude,
        b_total=backward_amplitude,
        a_subset=forward_amplitude_bwd_eigenmode,
        b_subset=backward_amplitude_bwd_eigenmode,
    )
    return jnp.real(s_forward), jnp.real(s_backward)


def eigenmode_poynting_flux(
    layer_solve_result: fmm.LayerSolveResult,
) -> jnp.ndarray:
    """Returns the total Poynting flux for each eigenmode.

    The result is equivalent to summing over the orders of the flux calculated
    by `amplitude_poynting_flux`, if the calculation is done for each eigenmode
    with a one-hot forward amplitude vector.

    Args:
        layer_solve_result: The results of the layer eigensolve.

    Returns:
        The per-eigenmode Poynting flux, with the same shape as the eigenvalues.
    """
    alpha_h = layer_solve_result.eigenvectors
    alpha_e = _poynting_flux_a_matrix(layer_solve_result)
    s_eigenmode = jnp.asarray(0.5) * (
        jnp.conj(alpha_e) * alpha_h + jnp.conj(alpha_h) * alpha_e
    )
    flux = jnp.sum(jnp.real(s_eigenmode), axis=-2)
    assert flux.shape == layer_solve_result.eigenvalues.shape
    return flux


def _poynting_flux_a_matrix(layer_solve_result: fmm.LayerSolveResult) -> jnp.ndarray:
    """Computes the `A` matrix from section 5.1 of [2012 Liu]."""
    q = layer_solve_result.eigenvalues
    phi = layer_solve_result.eigenvectors
    omega_script_k = layer_solve_result.omega_script_k_matrix
    angular_frequency = utils.angular_frequency_for_wavelength(
        layer_solve_result.wavelength
    )[..., jnp.newaxis]
    return (
        omega_script_k
        @ phi
        @ utils.diag(jnp.ones((), dtype=q.dtype) / (angular_frequency * q))
    )


def fields_from_wave_amplitudes(
    forward_amplitude: jnp.ndarray,
    backward_amplitude: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
) -> Tuple[
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
]:
    """Computes the electric and magnetic fields inside a layer.

    The calculation is for a batch of amplitudes, with the batch axis being the
    final axis. There can also be leading batch axes. Accordingly, amplitudes
    should have shape `(..., 2 * num_terms, num_amplitudes)`. The trailing batch
    dimension is preferred because it allows matrix-matrix multiplication instead
    of batched matrix-vector multiplication.

    Args:
        forward_amplitude: The amplitude of the forward-propagating waves.
        backward_amplitude: The amplitude of the backward-propagating waves,
            at the same location in space as the `forward_amplitude`.
        layer_solve_result: The results of the layer eigensolve.

    Returns:
        The electric and magnetic fields, `((ex, ey, ez), (hx, hy, hz))`.
    """
    _validate_amplitudes_shape(
        (forward_amplitude, backward_amplitude),
        num_terms=2 * layer_solve_result.expansion.num_terms,
    )

    # The matrix from equation 35 of [2012 Liu].
    matrix = field_conversion_matrix(layer_solve_result)

    # Obtain the transverse electric and magnetic fields.
    amplitudes = jnp.concatenate([forward_amplitude, backward_amplitude], axis=-2)

    fields = matrix @ amplitudes
    # The signs and ordering of fields is defined by equations 31 and 32 of [2012 Liu].
    negative_ey, ex, hx, hy = jnp.split(fields, 4, axis=-2)
    ey = -negative_ey

    # Compute the z-directed electric field, using equation 11 from [2012 Liu].
    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=layer_solve_result.in_plane_wavevector,
        primitive_lattice_vectors=layer_solve_result.primitive_lattice_vectors,
        expansion=layer_solve_result.expansion,
    )
    kx = transverse_wavevectors[..., 0, jnp.newaxis]
    ky = transverse_wavevectors[..., 1, jnp.newaxis]
    angular_frequency = utils.angular_frequency_for_wavelength(
        layer_solve_result.wavelength
    )
    angular_frequency = angular_frequency[..., jnp.newaxis, jnp.newaxis]

    # We use the the Fourier convolution matrix for the inverse of permittivity,
    # rather than inverting the Fourier convolution matrix of permittivity itself.
    # This improves convergence of the computed z-component of electric field.
    ez = (
        -layer_solve_result.inverse_z_permittivity_matrix
        @ (1j * kx * hy - 1j * ky * hx)
        / (1j * angular_frequency)
    )

    # Compute the z-directed magnetic field. The expression is similar to
    # equation 14 from [2012 Liu], but modified to allow for magnetic materials.
    hz = (
        layer_solve_result.inverse_z_permeability_matrix
        @ (1j * kx * ey - 1j * ky * ex)
        / (1j * angular_frequency)
    )

    assert ex.shape == ey.shape == ez.shape == hx.shape == hy.shape == hz.shape
    return (ex, ey, ez), (hx, hy, hz)


def field_conversion_matrix(layer_solve_result: fmm.LayerSolveResult) -> jnp.ndarray:
    """Returns the matrix which converts wave amplitudes to transverse fields."""
    # The matrix is from equation 35 of [2012 Liu].
    q = layer_solve_result.eigenvalues
    phi = layer_solve_result.eigenvectors
    omega_script_k = layer_solve_result.omega_script_k_matrix
    angular_frequency = utils.angular_frequency_for_wavelength(
        layer_solve_result.wavelength
    )[..., jnp.newaxis]

    # Note that there is a factor of `angular_frequency` in the denominator here, which
    # differs from equation 35 in [2012 Liu]. This is an error in that reference, and
    # the factor is actually present e.g. in equation 59.
    mat = (
        omega_script_k
        @ phi
        @ utils.diag(jnp.ones((), dtype=q.dtype) / (angular_frequency * q))
    )
    return jnp.block([[mat, -mat], [phi, phi]])


# -----------------------------------------------------------------------------
# Functions compute fields at grid locations in a single xy plane.
# -----------------------------------------------------------------------------


# Type for functions that compute fields at grid locations.
FieldsXYSliceFn = Callable[
    [
        Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # Ex, Ey, Ez Fourier amplitudes
        Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # Hx, Hy, Hz Fourier amplitudes
        fmm.LayerSolveResult,
    ],
    Tuple[
        Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # Ex, Ey, Ez
        Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # Hx, Hy, Hz
        Tuple[jnp.ndarray, jnp.ndarray],  # x, y
    ],
]


def fields_on_grid(
    electric_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    magnetic_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    layer_solve_result: fmm.LayerSolveResult,
    shape: Tuple[int, int],
    num_unit_cells: Tuple[int, int],
) -> Tuple[
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    Tuple[jnp.ndarray, jnp.ndarray],
]:
    """Transforms the fields from fourier representation to the grid.

    The fields within an array of unit cells is returned, with the number of
    cells in each direction given by `num_unit_cells`.

    The calculation is for a batch of fields, with the batch axis being the
    final axis. There can also be leading batch axes. Accordingly, fields
    should have shape `(..., 2 * num_terms, num_amplitudes)`. The trailing batch
    dimension is preferred because it allows matrix-matrix multiplication instead
    of batched matrix-vector multiplication.

    Args:
        electric_field: `(ex, ey, ez)` electric field Fourier amplitudes.
        magnetic_field: `(hx, hy, hz)` magnetic field Fourier amplitudes.
        layer_solve_result: The results of the layer eigensolve.
        shape: The shape of the grid.
        num_unit_cells: The number of unit cells along each direction.

    Returns:
        The electric field `(ex, ey, ez)`, magnetic field `(hx, hy, hz)`,
        and the grid coordinates `(x, y)`.
    """
    return _fields_on_grid(
        electric_field=electric_field,
        magnetic_field=magnetic_field,
        shape=shape,
        num_unit_cells=num_unit_cells,
        in_plane_wavevector=layer_solve_result.in_plane_wavevector,
        primitive_lattice_vectors=layer_solve_result.primitive_lattice_vectors,
        expansion=layer_solve_result.expansion,
    )


def _fields_on_grid(
    electric_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    magnetic_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    shape: Tuple[int, int],
    num_unit_cells: Tuple[int, int],
) -> Tuple[
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    Tuple[jnp.ndarray, jnp.ndarray],
]:
    """Transforms the fields from fourier representation to the grid."""
    _validate_amplitudes_shape(
        electric_field + magnetic_field,
        num_terms=expansion.num_terms,
    )
    x, y = basis.unit_cell_coordinates(
        primitive_lattice_vectors=primitive_lattice_vectors,
        shape=shape,
        num_unit_cells=num_unit_cells,
    )
    kx = in_plane_wavevector[..., 0, jnp.newaxis, jnp.newaxis]
    ky = in_plane_wavevector[..., 1, jnp.newaxis, jnp.newaxis]
    phase = jnp.exp(1j * (kx * x + ky * y))[..., jnp.newaxis]
    assert (
        x.shape[-2:]
        == y.shape[-2:]
        == (shape[0] * num_unit_cells[0], shape[1] * num_unit_cells[1])
    )

    def _field_on_grid(fourier_field):
        field = fft.ifft(fourier_field, expansion, shape, axis=-2)
        return jnp.tile(field, num_unit_cells + (1,))

    ex, ey, ez = electric_field
    grid_electric_field = (
        _field_on_grid(ex) * phase,
        _field_on_grid(ey) * phase,
        _field_on_grid(ez) * phase,
    )

    hx, hy, hz = magnetic_field
    grid_magnetic_field = (
        _field_on_grid(hx) * phase,
        _field_on_grid(hy) * phase,
        _field_on_grid(hz) * phase,
    )
    return grid_electric_field, grid_magnetic_field, (x, y)


def fields_on_coordinates(
    electric_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    magnetic_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    layer_solve_result: fmm.LayerSolveResult,
    x: jnp.ndarray,
    y: jnp.ndarray,
) -> Tuple[
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    Tuple[jnp.ndarray, jnp.ndarray],
]:
    """Computes the fields at specified coordinates.

    The calculation is for a batch of fields, with the batch axis being the
    final axis. There can also be leading batch axes. Accordingly, fields
    should have shape `(..., 2 * num_terms, num_amplitudes)`. The trailing batch
    dimension is preferred because it allows matrix-matrix multiplication instead
    of batched matrix-vector multiplication.

    Args:
        electric_field: `(ex, ey, ez)` electric field Fourier amplitudes.
        magnetic_field: `(hx, hy, hz)` magnetic field Fourier amplitudes.
        layer_solve_result: The results of the layer eigensolve.
        x: The x-coordinates where the fields are sought.
        y: The y-coordinates where the fields are sought, with shape matching
            that of `x`.

    Returns:
        The electric field `(ex, ey, ez)`, magnetic field `(hx, hy, hz)`,
        and the grid coordinates `(x, y)`. The field arrays each have shape
        `batch_shape + coordinates_shape + (num_amplitudes)`.
    """
    _validate_amplitudes_shape(
        electric_field + magnetic_field,
        num_terms=layer_solve_result.expansion.num_terms,
    )
    assert x.shape == y.shape

    coordinates_shape = x.shape
    ex_shape = electric_field[0].shape
    field_shape = ex_shape[:-2] + coordinates_shape + (ex_shape[-1],)

    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=layer_solve_result.in_plane_wavevector,
        primitive_lattice_vectors=layer_solve_result.primitive_lattice_vectors,
        expansion=layer_solve_result.expansion,
    )
    kx = transverse_wavevectors[..., 0, jnp.newaxis]
    ky = transverse_wavevectors[..., 1, jnp.newaxis]

    def _field_at_coordinates(fourier_field):
        field = (
            fourier_field[..., jnp.newaxis, :]
            * jnp.exp(1j * (kx * x.flatten() + ky * y.flatten()))[..., jnp.newaxis]
        )
        field = jnp.sum(field, axis=-3)
        return field.reshape(field_shape)

    ex, ey, ez = electric_field
    grid_electric_field = (
        _field_at_coordinates(ex),
        _field_at_coordinates(ey),
        _field_at_coordinates(ez),
    )

    hx, hy, hz = magnetic_field
    grid_magnetic_field = (
        _field_at_coordinates(hx),
        _field_at_coordinates(hy),
        _field_at_coordinates(hz),
    )
    return grid_electric_field, grid_magnetic_field, (x, y)


# -----------------------------------------------------------------------------
# Functions to wave amplitudes inside a stack of layers.
# -----------------------------------------------------------------------------


def stack_amplitudes_interior(
    s_matrices_interior: Tuple[
        Tuple[scattering.ScatteringMatrix, scattering.ScatteringMatrix], ...
    ],
    forward_amplitude_0_start: jnp.ndarray,
    backward_amplitude_N_end: jnp.ndarray,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], ...]:
    """Computes the wave amplitudes at interior layers within a stack.

    The calculation is for a batch of amplitudes, with the batch axis being the
    final axis. There can also be leading batch axes. Accordingly, amplitudes
    should have shape `(..., 2 * num_terms, num_amplitudes)`. The trailing batch
    dimension is preferred because it allows matrix-matrix multiplication instead
    of batched matrix-vector multiplication.

    Args:
        s_matrices_interior: The scattering matrices for the substacks before
            and after each layer, as computed by `stack_s_matrices_interior`.
        forward_amplitude_0_start: The forward-propagating wave amplitude at the
            start of the first layer of the stack.
        backward_amplitude_N_end: The backward-propagating wave amplitude at the
            end of the last layer of the stack.

    Returns:
        The forward- and backward-propagating wave amplitude for each layer,
        defined at the start and end of each layer, respectively.
    """
    return tuple(
        [
            amplitudes_interior(
                s_matrix_before=s_matrix_before,
                s_matrix_after=s_matrix_after,
                forward_amplitude_0_start=forward_amplitude_0_start,
                backward_amplitude_N_end=backward_amplitude_N_end,
            )
            for (s_matrix_before, s_matrix_after) in s_matrices_interior
        ]
    )


def stack_amplitudes_interior_with_source(
    s_matrices_interior_before_source: Tuple[
        Tuple[scattering.ScatteringMatrix, scattering.ScatteringMatrix], ...
    ],
    s_matrices_interior_after_source: Tuple[
        Tuple[scattering.ScatteringMatrix, scattering.ScatteringMatrix], ...
    ],
    backward_amplitude_before_end: jnp.ndarray,
    forward_amplitude_after_start: jnp.ndarray,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], ...]:
    """Computes the wave amplitudes in the case of an internal source.

    Args:
        s_matrices_interior_before_source: The interior scattering matrices for
            the layer substrack before the source, as computed by
            `stack_s_matrices_interior`.
        s_matrices_interior_after_source: The interior scattering matrices for
            the layer substack after the source.
        backward_amplitude_before_end: The backward-going wave amplitude at the
            end of the layer before the source.
        forward_amplitude_after_start: The forward-going wave amplitude at the
            start of the layer after the source.

    Returns:
        The forward- and backward-propagating wave amplitude for each layer,
        defined at the start and end of each layer, respectively.
    """
    amplitudes_before = stack_amplitudes_interior(
        s_matrices_interior=s_matrices_interior_before_source,
        forward_amplitude_0_start=jnp.zeros_like(backward_amplitude_before_end),
        backward_amplitude_N_end=backward_amplitude_before_end,
    )
    amplitudes_after = stack_amplitudes_interior(
        s_matrices_interior=s_matrices_interior_after_source,
        forward_amplitude_0_start=forward_amplitude_after_start,
        backward_amplitude_N_end=jnp.zeros_like(forward_amplitude_after_start),
    )
    return amplitudes_before + amplitudes_after


def amplitudes_interior(
    s_matrix_before: scattering.ScatteringMatrix,
    s_matrix_after: scattering.ScatteringMatrix,
    forward_amplitude_0_start: jnp.ndarray,
    backward_amplitude_N_end: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the wave amplitudes at an interior layer within a stack.

    The calculation is for a batch of amplitudes, with the batch axis being the
    final axis. There can also be leading batch axes. Accordingly, amplitudes
    should have shape `(..., 2 * num_terms, num_amplitudes)`. The trailing batch
    dimension is preferred because it allows matrix-matrix multiplication instead
    of batched matrix-vector multiplication.

    Args:
        s_matrix_before: The scattering matrix for the substack before the layer.
        s_matrix_after: The scattering matrix for the substack after the layer.
        forward_amplitude_0_start: The forward-propagating wave amplitude at the
            start of the first layer of the stack.
        backward_amplitude_N_end: The backward-propagating wave amplitude at the
            end of the last layer of the stack.

    Returns:
        The forward- and backward-propagating wave amplitude in the layer, defined
        at layer start and end, respectively.
    """
    _validate_amplitudes_shape(
        (forward_amplitude_0_start, backward_amplitude_N_end),
        num_terms=2 * s_matrix_before.start_layer_solve_result.expansion.num_terms,
    )
    num = forward_amplitude_0_start.shape[-2]

    # Solve for the interior forward-going wave amplitude at the start of the layer.
    forward_rhs = (
        s_matrix_before.s11 @ forward_amplitude_0_start
        + s_matrix_before.s12 @ s_matrix_after.s22 @ backward_amplitude_N_end
    )
    forward_mat = jnp.eye(num) - s_matrix_before.s12 @ s_matrix_after.s21
    forward_amplitude_i_start = jnp.linalg.solve(forward_mat, forward_rhs)

    # The interior backward-going wave amplitude can now be directly computed.
    backward_amplitude_i_end = (
        s_matrix_after.s21 @ forward_amplitude_i_start
        + s_matrix_after.s22 @ backward_amplitude_N_end
    )
    return forward_amplitude_i_start, backward_amplitude_i_end


# -----------------------------------------------------------------------------
# Functions to compute fields on the 3D real-space grid.
# -----------------------------------------------------------------------------


def stack_fields_3d_auto_grid(
    amplitudes_interior: Sequence[Tuple[jnp.ndarray, jnp.ndarray]],
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    grid_spacing: float,
    num_unit_cells: Tuple[int, int],
) -> Tuple[jnp.ndarray, jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:
    """Computes the three-dimensional fields in a stack on the real-space grid.

    The grid is automatically determined from the layer dimensions and the resolution.

    Args:
        amplitudes_interior: The forward- and backward-propagating wave amplitude
            for each layer, defined at the start and end of each layer, respectively.
        layer_solve_results: The results of the layer eigensolve for each layer.
        layer_thicknesses: The thickness of each layer.
        grid_spacing: The approximate spacing of gridpoints on which the field is
            computed. The actual grid spacing is modified to align with the layer
            and unit cell boundaries.
        num_unit_cells: The number of unit cells along each direction.

    Returns:
        The electric and magnetic fields and grid coordinates, `(ef, hf, (x, y, z))`.
    """
    primitive_lattice_vectors = layer_solve_results[0].primitive_lattice_vectors
    grid_shape = (
        int(jnp.round(jnp.linalg.norm(primitive_lattice_vectors.u) / grid_spacing)),
        int(jnp.round(jnp.linalg.norm(primitive_lattice_vectors.v) / grid_spacing)),
    )

    layer_znum = tuple([int(jnp.round(t / grid_spacing)) for t in layer_thicknesses])

    return stack_fields_3d(
        amplitudes_interior=amplitudes_interior,
        layer_solve_results=layer_solve_results,
        layer_thicknesses=layer_thicknesses,
        layer_znum=layer_znum,
        grid_shape=grid_shape,
        num_unit_cells=num_unit_cells,
    )


def stack_fields_3d(
    amplitudes_interior: Sequence[Tuple[jnp.ndarray, jnp.ndarray]],
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    layer_znum: Sequence[int],
    grid_shape: Tuple[int, int],
    num_unit_cells: Tuple[int, int],
) -> Tuple[jnp.ndarray, jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:
    """Computes the three-dimensional fields in a stack on the real-space grid.

    Args:
        amplitudes_interior: The forward- and backward-propagating wave amplitude
            for each layer, defined at the start and end of each layer, respectively.
        layer_solve_results: The results of the layer eigensolve for each layer.
        layer_thicknesses: The thickness of each layer.
        layer_znum: The number of gridpoints in the z-direction for each layer.
        grid_shape: The shape of the xy real-space grid.
        num_unit_cells: The number of unit cells along each direction.

    Returns:
        The electric and magnetic fields and grid coordinates, `(ef, hf, (x, y, z))`.
    """
    return _stack_fields_3d(
        amplitudes_interior=amplitudes_interior,
        layer_solve_results=layer_solve_results,
        layer_thicknesses=layer_thicknesses,
        layer_znum=layer_znum,
        fields_xy_slice_fn=functools.partial(
            fields_on_grid,
            shape=grid_shape,
            num_unit_cells=num_unit_cells,
        ),
    )


def stack_fields_3d_on_coordinates(
    amplitudes_interior: Sequence[Tuple[jnp.ndarray, jnp.ndarray]],
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    layer_znum: Sequence[int],
    x: jnp.ndarray,
    y: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:
    """Computes the three-dimensional fields in a stack at specified coordinates.

    This function may be significantly faster than `stack_fields_3d` in cases where
    fields in the full simulation domain are not required.

    Args:
        amplitudes_interior: The forward- and backward-propagating wave amplitude
            for each layer, defined at the start and end of each layer, respectively.
        layer_solve_results: The results of the layer eigensolve for each layer.
        layer_thicknesses: The thickness of each layer.
        layer_znum: The number of gridpoints in the z-direction for each layer.
        x: The x-coordinates where the fields are sought.
        y: The y-coordinates where the fields are sought, with shape matching
            that of `x`.

    Returns:
        The electric and magnetic fields and grid coordinates, `(ef, hf, (x, y, z))`.
    """
    return _stack_fields_3d(
        amplitudes_interior=amplitudes_interior,
        layer_solve_results=layer_solve_results,
        layer_thicknesses=layer_thicknesses,
        layer_znum=layer_znum,
        fields_xy_slice_fn=functools.partial(fields_on_coordinates, x=x, y=y),
    )


def layer_fields_3d(
    forward_amplitude_start: jnp.ndarray,
    backward_amplitude_end: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
    layer_thickness: jnp.ndarray,
    layer_znum: int,
    grid_shape: Tuple[int, int],
    num_unit_cells: Tuple[int, int],
) -> Tuple[jnp.ndarray, jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:
    """Computes the three-dimensional fields in a layer on the real-space grid.

    Args:
        forward_amplitude_start: The forward-going wave amplitudes, defined at the
            start of the layer.
        backward_amplitude_end: The backward-going wave amplitudes, defined at the
            end of the layer.
        layer_solve_result: The results of the layer eigensolve.
        layer_thickness: The layer thickness.
        layer_znum: The number of gridpoints in the z-direction for the layer.
        grid_shape: The shape of the xy real-space grid.
        num_unit_cells: The number of unit cells along each direction.

    Returns:
        The electric and magnetic fields and grid coordinates, `(ef, hf, (x, y, z))`.
    """
    return _layer_fields_3d(
        forward_amplitude_start=forward_amplitude_start,
        backward_amplitude_end=backward_amplitude_end,
        layer_solve_result=layer_solve_result,
        layer_thickness=layer_thickness,
        layer_znum=layer_znum,
        fields_xy_slice_fn=functools.partial(
            fields_on_grid,
            shape=grid_shape,
            num_unit_cells=num_unit_cells,
        ),
    )


def layer_fields_3d_on_coordinates(
    forward_amplitude_start: jnp.ndarray,
    backward_amplitude_end: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
    layer_thickness: jnp.ndarray,
    layer_znum: int,
    x: jnp.ndarray,
    y: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:
    """Computes the three-dimensional fields in a layer at specified coordinates

    This function may be significantly faster than `layer_fields_3d` in cases where
    fields in the full simulation domain are not required.

    Args:
        forward_amplitude_start: The forward-going wave amplitudes, defined at the
            start of the layer.
        backward_amplitude_end: The backward-going wave amplitudes, defined at the
            end of the layer.
        layer_solve_result: The results of the layer eigensolve.
        layer_thickness: The layer thickness.
        layer_znum: The number of gridpoints in the z-direction for the layer.
        x: The x-coordinates where the fields are sought.
        y: The y-coordinates where the fields are sought, with shape matching
            that of `x`.

    Returns:
        The electric and magnetic fields and grid coordinates, `(ef, hf, (x, y, z))`.
    """
    return _layer_fields_3d(
        forward_amplitude_start=forward_amplitude_start,
        backward_amplitude_end=backward_amplitude_end,
        layer_solve_result=layer_solve_result,
        layer_thickness=layer_thickness,
        layer_znum=layer_znum,
        fields_xy_slice_fn=functools.partial(fields_on_coordinates, x=x, y=y),
    )


def _stack_fields_3d(
    amplitudes_interior: Sequence[Tuple[jnp.ndarray, jnp.ndarray]],
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    layer_znum: Sequence[int],
    fields_xy_slice_fn: FieldsXYSliceFn,
) -> Tuple[jnp.ndarray, jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:
    """Computes the three-dimensional fields in a stack on the real-space grid.

    Args:
        amplitudes_interior: The forward- and backward-propagating wave amplitude
            for each layer, defined at the start and end of each layer, respectively.
        layer_solve_results: The results of the layer eigensolve for each layer.
        layer_thicknesses: The thickness of each layer.
        layer_znum: The number of gridpoints in the z-direction for each layer.
        fields_xy_slice_fn: Computes the fields for each xy slice given the field
            Fourier amplitudes and layer solve result.

    Returns:
        The electric and magnetic fields and grid coordinates, `(ef, hf, (x, y, z))`.
    """
    _validate_matching_lengths(
        amplitudes_interior, layer_solve_results, layer_thicknesses, layer_znum
    )

    z0 = jnp.zeros(())
    zs = []
    efields = []
    hfields = []
    for layer_amplitudes, layer_solve_result, layer_thickness, znum in zip(
        amplitudes_interior,
        layer_solve_results,
        layer_thicknesses,
        layer_znum,
        strict=True,
    ):
        forward_amplitude_start, backward_amplitude_end = layer_amplitudes
        assert forward_amplitude_start.shape == backward_amplitude_end.shape
        eg, hg, (x, y, z_offset) = _layer_fields_3d(
            forward_amplitude_start=forward_amplitude_start,
            backward_amplitude_end=backward_amplitude_end,
            layer_solve_result=layer_solve_result,
            layer_thickness=layer_thickness,
            layer_znum=znum,
            fields_xy_slice_fn=fields_xy_slice_fn,
        )
        efields.append(eg)
        hfields.append(hg)
        zs.append(z_offset + z0)
        z0 += layer_thickness

    merged_efields = jnp.concatenate(efields, axis=-2)
    merged_hfields = jnp.concatenate(hfields, axis=-2)
    z = jnp.concatenate(zs)
    return merged_efields, merged_hfields, (x, y, z)


def _layer_fields_3d(
    forward_amplitude_start: jnp.ndarray,
    backward_amplitude_end: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
    layer_thickness: jnp.ndarray,
    layer_znum: int,
    fields_xy_slice_fn: FieldsXYSliceFn,
) -> Tuple[jnp.ndarray, jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:
    """Computes the three-dimensional fields in a layer on the real-space grid.

    Args:
        forward_amplitude_start: The forward-going wave amplitudes, defined at the
            start of the layer.
        backward_amplitude_end: The backward-going wave amplitudes, defined at the
            end of the layer.
        layer_solve_result: The results of the layer eigensolve.
        layer_thickness: The layer thickness.
        layer_znum: The number of gridpoints in the z-direction for the layer.
        fields_xy_slice_fn: Computes the fields for each xy slice given the field
            Fourier amplitudes and layer solve result.

    Returns:
        The electric and magnetic fields and grid coordinates, `(ef, hf, (x, y, z))`.
    """

    assert forward_amplitude_start.shape == backward_amplitude_end.shape

    z_offset = jnp.linspace(0, layer_thickness, layer_znum)

    # Add a trailing batch dimension to `z_offset`, which matches the trailing
    # batch dimension for the amplitudes.
    amplitude_batch_size = forward_amplitude_start.shape[-1]
    z_offset_broadcast = jnp.broadcast_to(
        z_offset[:, jnp.newaxis], (layer_znum, amplitude_batch_size)
    )

    # Add a batch dimension to the amplitudes, for the offset.
    forward_amplitude_start = jnp.broadcast_to(
        forward_amplitude_start[..., jnp.newaxis, :],
        forward_amplitude_start.shape[:-1] + (layer_znum, amplitude_batch_size),
    )
    backward_amplitude_end = jnp.broadcast_to(
        backward_amplitude_end[..., jnp.newaxis, :],
        backward_amplitude_end.shape[:-1] + (layer_znum, amplitude_batch_size),
    )

    # Flatten the trailing batch dimensions for offset and amplitudes.
    z_offset_broadcast = z_offset_broadcast.flatten()
    forward_amplitude_start = forward_amplitude_start.reshape(
        forward_amplitude_start.shape[:-2] + (-1,)
    )
    backward_amplitude_end = backward_amplitude_end.reshape(
        backward_amplitude_end.shape[:-2] + (-1,)
    )

    # Compute the amplitudes everywhere in the layer. This creates a batch of
    # amplitudes. Since each of those are associated with the same layer solve
    # result, we ensure that the batch axis is the final axis, since this is
    # the "fast" batch axis for field computations.
    forward_amplitude, backward_amplitude = colocate_amplitudes(
        forward_amplitude_start,
        backward_amplitude_end,
        z_offset=z_offset_broadcast,
        layer_solve_result=layer_solve_result,
        layer_thickness=layer_thickness,
    )
    assert forward_amplitude.shape[-1] == z_offset_broadcast.size

    ef, hf = fields_from_wave_amplitudes(
        forward_amplitude=forward_amplitude,
        backward_amplitude=backward_amplitude,
        layer_solve_result=layer_solve_result,
    )
    eg_tuple, hg_tuple, (x, y) = fields_xy_slice_fn(
        electric_field=ef,  # type: ignore[call-arg]
        magnetic_field=hf,  # type: ignore[call-arg]
        layer_solve_result=layer_solve_result,  # type: ignore[call-arg]
    )
    eg = jnp.asarray(eg_tuple)
    hg = jnp.asarray(hg_tuple)

    # Restore the original amplitude batch dimension.
    eg = jnp.reshape(eg, eg.shape[:-1] + (layer_znum, amplitude_batch_size))
    hg = jnp.reshape(hg, hg.shape[:-1] + (layer_znum, amplitude_batch_size))

    return eg, hg, (x, y, z_offset)


def _validate_matching_lengths(*sequences: Sequence) -> None:
    """Validates that all of `args` have matching length."""
    lengths = [len(s) for s in sequences]
    if not all([length == lengths[0] for length in lengths]):
        raise ValueError(f"Encountered incompatible lengths, got lengths of {lengths}")



"""Functions related to layer eigenmode calculation for the FMM algorithm.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import dataclasses
import enum
import functools
from typing import Callable, Optional, Tuple

import jax.numpy as jnp
from jax import tree_util

from fmmax import basis, fft, fmm_matrices, utils, vector

# xx, xy, yx, yy, and zz components of permittivity or permeability.
TensorComponents = Tuple[
    jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray
]
VectorFn = Callable[
    [jnp.ndarray, basis.Expansion, basis.LatticeVectors],
    Tuple[jnp.ndarray, jnp.ndarray],
]


@enum.unique
class Formulation(enum.Enum):
    """Enumerates supported Fourier modal method formulations."""

    FFT = "fft"
    JONES_DIRECT = vector.JONES_DIRECT
    JONES = vector.JONES
    NORMAL = vector.NORMAL
    POL = vector.POL
    JONES_DIRECT_FOURIER = vector.JONES_DIRECT_FOURIER
    JONES_FOURIER = vector.JONES_FOURIER
    NORMAL_FOURIER = vector.NORMAL_FOURIER
    POL_FOURIER = vector.POL_FOURIER


def eigensolve_isotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation | VectorFn,
) -> "LayerSolveResult":
    """Performs the eigensolve for a layer with isotropic permittivity.

    This function performs either a uniform-layer or patterned-layer eigensolve,
    depending on the shape of the trailing dimensions of a given layer permittivity.
    When the final two dimensions have shape `(1, 1)`, the layer is treated as
    uniform. Otherwise, it is patterned.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The permittivity array.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used, or a callable which computes
            the tangent vector field for a custom vector FMM formulation.

    Returns:
        The `LayerSolveResult`.
    """
    if permittivity.shape[-2:] == (1, 1):
        _eigensolve_fn = _eigensolve_uniform_isotropic_media
    else:
        _eigensolve_fn = functools.partial(
            _eigensolve_patterned_isotropic_media, formulation=formulation
        )

    return _eigensolve_fn(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        permittivity=permittivity,
        expansion=expansion,
    )


def eigensolve_anisotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity_xx: jnp.ndarray,
    permittivity_xy: jnp.ndarray,
    permittivity_yx: jnp.ndarray,
    permittivity_yy: jnp.ndarray,
    permittivity_zz: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation | VectorFn,
) -> "LayerSolveResult":
    """Performs the eigensolve for a layer with anisotropic permittivity.

    This function performs either a uniform-layer or patterned-layer eigensolve,
    depending on the shape of the trailing dimensions of a given layer permittivity.
    When the final two dimensions have shape `(1, 1)`, the layer is treated as
    uniform. Otherwise, it is patterned.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity_xx: The xx-component of the permittivity tensor, with
            shape `(..., nx, ny)`.
        permittivity_xy: The xy-component of the permittivity tensor.
        permittivity_yx: The yx-component of the permittivity tensor.
        permittivity_yy: The yy-component of the permittivity tensor.
        permittivity_zz: The zz-component of the permittivity tensor.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used, or a callable which computes
            the tangent vector field for a custom vector FMM formulation.

    Returns:
        The `LayerSolveResult`.
    """
    return eigensolve_general_anisotropic_media(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        permittivity_xx=permittivity_xx,
        permittivity_xy=permittivity_xy,
        permittivity_yx=permittivity_yx,
        permittivity_yy=permittivity_yy,
        permittivity_zz=permittivity_zz,
        permeability_xx=jnp.ones_like(permittivity_xx),
        permeability_xy=jnp.zeros_like(permittivity_xx),
        permeability_yx=jnp.zeros_like(permittivity_xx),
        permeability_yy=jnp.ones_like(permittivity_xx),
        permeability_zz=jnp.ones_like(permittivity_xx),
        expansion=expansion,
        formulation=formulation,
        vector_field_source=None,
    )


def eigensolve_general_anisotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity_xx: jnp.ndarray,
    permittivity_xy: jnp.ndarray,
    permittivity_yx: jnp.ndarray,
    permittivity_yy: jnp.ndarray,
    permittivity_zz: jnp.ndarray,
    permeability_xx: jnp.ndarray,
    permeability_xy: jnp.ndarray,
    permeability_yx: jnp.ndarray,
    permeability_yy: jnp.ndarray,
    permeability_zz: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation | VectorFn,
    vector_field_source: Optional[jnp.ndarray] = None,
) -> "LayerSolveResult":
    """Performs the eigensolve for a general anistropic layer.

    Here, "general" refers to the fact that the layer material can be magnetic, i.e.
    the permeability and permittivity can be specified.

    This function performs either a uniform-layer or patterned-layer eigensolve,
    depending on the shape of the trailing dimensions of a given layer permittivity.
    When the final two dimensions have shape `(1, 1)`, the layer is treated as
    uniform. Otherwise, it is patterned.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity_xx: The xx-component of the permittivity tensor, with
            shape `(..., nx, ny)`.
        permittivity_xy: The xy-component of the permittivity tensor.
        permittivity_yx: The yx-component of the permittivity tensor.
        permittivity_yy: The yy-component of the permittivity tensor.
        permittivity_zz: The zz-component of the permittivity tensor.
        permeability_xx: The xx-component of the permeability tensor.
        permeability_xy: The xy-component of the permeability tensor.
        permeability_yx: The yx-component of the permeability tensor.
        permeability_yy: The yy-component of the permeability tensor.
        permeability_zz: The zz-component of the permeability tensor.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used, or a callable which computes
            the tangent vector field for a custom vector FMM formulation.
        vector_field_source: Optional array used to calculate the vector field for
            vector formulations of the FMM. If not specified, `(permittivity_xx +
            permittivity_yy) / 2` is used. Ignored for the `FFT` formulation. Should
            have shape matching the permittivities and permeabilities.

    Returns:
        The `LayerSolveResult`.
    """
    if permittivity_xx.shape[-2:] == (1, 1):
        _eigensolve_fn = _eigensolve_uniform_general_anisotropic_media
    else:
        if vector_field_source is None:
            vector_field_source = (permittivity_xx + permittivity_yy) / 2
        _eigensolve_fn = functools.partial(
            _eigensolve_patterned_general_anisotropic_media,
            formulation=formulation,
            vector_field_source=vector_field_source,
        )

    return _eigensolve_fn(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        permittivities=(
            permittivity_xx,
            permittivity_xy,
            permittivity_yx,
            permittivity_yy,
            permittivity_zz,
        ),
        permeabilities=(
            permeability_xx,
            permeability_xy,
            permeability_yx,
            permeability_yy,
            permeability_zz,
        ),
        expansion=expansion,
    )


@dataclasses.dataclass
class LayerSolveResult:
    """Stores the result of a layer eigensolve.

    This eigenvalue problem is specified in equation 28 of [2012 Liu].

    Attributes:
        wavelength: The wavelength for the solve.
        in_plane_wavevector: The in-plane wavevector for the solve.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        expansion: The expansion used for the eigensolve.
        eigenvalues: The layer eigenvalues.
        eigenvectors: The layer eigenvectors.
        z_permittivity_matrix: The fourier-transformed zz-component of permittivity.
        inverse_z_permittivity_matrix: The fourier-transformed inverse of zz-component
            of permittivity.
        z_permeability_matrix: The fourier-transformed zz-component of permeability.
        inverse_z_permeability_matrix: The fourier-transformed inverse of zz-component
            of permeability.
        transverse_permeability_matrix: The transverse permeability matrix, needed to
            calculate the omega-script-k matrix from equation 26 of [2012 Liu]. This
            is needed to generate the layer scattering matrix.
        tangent_vector_field: The tangent vector field `(tx, ty)` used to compute the
            transverse permittivity matrix, if a vector FMM formulation is used. If
            the `FFT` formulation is used, the vector field is `None`.
    """

    wavelength: jnp.ndarray
    in_plane_wavevector: jnp.ndarray
    primitive_lattice_vectors: basis.LatticeVectors
    expansion: basis.Expansion
    eigenvalues: jnp.ndarray
    eigenvectors: jnp.ndarray
    z_permittivity_matrix: jnp.ndarray
    inverse_z_permittivity_matrix: jnp.ndarray
    z_permeability_matrix: jnp.ndarray
    inverse_z_permeability_matrix: jnp.ndarray
    transverse_permeability_matrix: jnp.ndarray
    tangent_vector_field: Optional[Tuple[jnp.ndarray, jnp.ndarray]]

    @property
    def batch_shape(self) -> Tuple[int, ...]:
        return self.eigenvectors.shape[:-2]

    @property
    def omega_script_k_matrix(self):
        """Compute omega-script-k matrix from equation 26 of [2012 Liu]."""
        return fmm_matrices.omega_script_k_matrix_patterned(
            wavelength=self.wavelength,
            z_permittivity_matrix=self.z_permittivity_matrix,
            transverse_permeability_matrix=self.transverse_permeability_matrix,
            transverse_wavevectors=basis.transverse_wavevectors(
                in_plane_wavevector=self.in_plane_wavevector,
                primitive_lattice_vectors=self.primitive_lattice_vectors,
                expansion=self.expansion,
            ),
        )

    def __post_init__(self) -> None:
        """Validates shapes of the `LayerSolveResult` attributes."""

        def _incompatible(arr: jnp.ndarray, reference_shape: Tuple[int, ...]) -> bool:
            ndim_mismatch = arr.ndim != len(reference_shape)
            batch_compatible = utils.batch_compatible_shapes(arr.shape, reference_shape)
            return ndim_mismatch or not batch_compatible

        if _incompatible(self.wavelength, self.batch_shape):
            raise ValueError(
                f"`wavelength` must have compatible batch shape, but got shape "
                f"{self.wavelength.shape} when `eigenvectors` shape is "
                f"{self.eigenvectors.shape}."
            )
        if _incompatible(self.in_plane_wavevector, self.batch_shape + (2,)):
            raise ValueError(
                f"`in_plane_wavevector` must have compatible batch shape, but got "
                f"shape {self.in_plane_wavevector.shape} when `eigenvectors` shape is "
                f"{self.eigenvectors.shape}."
            )
        if _incompatible(self.primitive_lattice_vectors.u, self.batch_shape + (2,)):
            raise ValueError(
                f"`primitive_lattice_vectors.u` must have compatible batch shape, but "
                f"got shape {self.primitive_lattice_vectors.u.shape} when "
                f"`eigenvectors` shape is {self.eigenvectors.shape}."
            )
        if _incompatible(self.primitive_lattice_vectors.v, self.batch_shape + (2,)):
            raise ValueError(
                f"`primitive_lattice_vectors.v` must have compatible batch shape, but "
                f"got shape {self.primitive_lattice_vectors.v.shape} when "
                f"`eigenvectors` shape is {self.eigenvectors.shape}."
            )
        if self.expansion.num_terms * 2 != self.eigenvectors.shape[-1]:
            raise ValueError(
                f"`eigenvectors` must have shape compatible with `expansion.num_terms`,"
                f" but got shape {self.eigenvectors.shape} when `num_terms` shape is "
                f"{self.expansion.num_terms}."
            )
        if self.eigenvalues.shape != self.eigenvectors.shape[:-1]:
            raise ValueError(
                f"`eigenvalues` must have compatible shape, but got shape "
                f"{self.eigenvalues.shape} when `eigenvectors` shape is "
                f"{self.eigenvectors.shape}."
            )

        expected_matrix_shape = self.batch_shape + (self.expansion.num_terms,) * 2
        if _incompatible(self.inverse_z_permittivity_matrix, expected_matrix_shape):
            raise ValueError(
                f"`inverse_z_permittivity_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes "
                f"{self.inverse_z_permittivity_matrix.shape} "
                f"and {self.eigenvectors.shape}."
            )
        if _incompatible(self.z_permittivity_matrix, expected_matrix_shape):
            raise ValueError(
                f"`z_permittivity_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes {self.z_permittivity_matrix.shape} "
                f"and {self.eigenvectors.shape}."
            )
        if _incompatible(self.inverse_z_permeability_matrix, expected_matrix_shape):
            raise ValueError(
                f"`inverse_z_permeability_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes "
                f"{self.inverse_z_permeability_matrix.shape} "
                f"and {self.eigenvectors.shape}."
            )
        if _incompatible(self.z_permeability_matrix, expected_matrix_shape):
            raise ValueError(
                f"`z_permeability_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes {self.z_permeability_matrix.shape} "
                f"and {self.eigenvectors.shape}."
            )
        if _incompatible(self.transverse_permeability_matrix, self.eigenvectors.shape):
            raise ValueError(
                f"`transverse_permeability_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes "
                f"{self.transverse_permeability_matrix.shape} and "
                f"s{self.eigenvectors.shape}."
            )

        if self.tangent_vector_field is not None and (
            self.tangent_vector_field[0].ndim != self.eigenvectors.ndim
        ):
            raise ValueError(
                f"`tangent_vector_field` must have ndim compatible with "
                f"`eigenvectors`, but got shapes {self.tangent_vector_field[0]} and "
                f"{self.eigenvectors}."
            )


# -----------------------------------------------------------------------------
# Eigensolves for specific cases, e.g. uniform isotropic, anisotropic, etc.
# -----------------------------------------------------------------------------


def _eigensolve_uniform_isotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
) -> LayerSolveResult:
    r"""Returns the the results of a uniform isotropic layer eigensolve.

    The layer is uniform and isotropic, in the sense that the permittivity does not
    vary spatially and has no orientation dependence. In this case, the eigenvalues
    and eigenvectors can be calculated analytically.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The scalar permittivity for the layer, with shape `(..., 1, 1)`.
        expansion: The field expansion to be used.

    Returns:
        The `LayerSolveResult`.
    """
    (
        wavelength,
        in_plane_wavevector,
        primitive_lattice_vectors,
        (permittivity,),
    ) = _validate_and_broadcast(
        wavelength, in_plane_wavevector, primitive_lattice_vectors, permittivity
    )
    if permittivity.shape[-2:] != (1, 1):
        raise ValueError(
            f"Trailing axes of `permittivity` must have shape (1, 1) but got a shape "
            f"of {permittivity.shape}."
        )

    batch_shape = jnp.broadcast_shapes(
        wavelength.shape, in_plane_wavevector.shape[:-1], permittivity.shape[:-2]
    )

    num_eigenvalues = 2 * expansion.num_terms
    permittivity = jnp.squeeze(permittivity, axis=(-2, -1))

    # Transverse wavevectors are the `k + G` from equation 5 of [2012 Liu].
    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )

    # In uniform media, the eigenvectors are just the plane waves.
    eigenvectors = jnp.broadcast_to(
        jnp.eye(num_eigenvalues, dtype=complex),
        shape=batch_shape + (num_eigenvalues, num_eigenvalues),
    )

    angular_frequency = utils.angular_frequency_for_wavelength(wavelength)
    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]
    eigenvalues = jnp.sqrt(
        (
            permittivity[..., jnp.newaxis] * angular_frequency[..., jnp.newaxis] ** 2
            - kx**2
            - ky**2
        ).astype(complex)
    )
    eigenvalues = _select_eigenvalues_sign(eigenvalues)
    eigenvalues = jnp.tile(eigenvalues, 2)

    batch_shape = eigenvalues.shape[:-1]
    inverse_z_permittivity_matrix = jnp.broadcast_to(
        1 / permittivity[..., jnp.newaxis], batch_shape + (expansion.num_terms,)
    )
    inverse_z_permittivity_matrix = utils.diag(inverse_z_permittivity_matrix)
    z_permittivity_matrix = jnp.broadcast_to(
        permittivity[..., jnp.newaxis], batch_shape + (expansion.num_terms,)
    )
    z_permittivity_matrix = utils.diag(z_permittivity_matrix)
    z_permeability_matrix = utils.diag(
        jnp.ones(
            z_permittivity_matrix.shape[:-1],
            dtype=z_permittivity_matrix.dtype,
        )
    )
    return LayerSolveResult(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        eigenvalues=eigenvalues,
        eigenvectors=eigenvectors,
        z_permittivity_matrix=z_permittivity_matrix,
        inverse_z_permittivity_matrix=inverse_z_permittivity_matrix,
        z_permeability_matrix=z_permeability_matrix,
        inverse_z_permeability_matrix=z_permeability_matrix,
        transverse_permeability_matrix=utils.diag(jnp.ones_like(eigenvalues)),
        tangent_vector_field=None,
    )


def _eigensolve_patterned_isotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation | VectorFn,
) -> LayerSolveResult:
    r"""Returns the results of a patterned isotropic layer eigensolve.

    The layer is patterned and isotropic, in the sense that the permittivity varies
    spatially and has no orientation dependence.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The permittivity array.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used, or a callable which computes
            the tangent vector field for a custom vector FMM formulation.

    Returns:
        The `LayerSolveResult`.
    """
    (
        wavelength,
        in_plane_wavevector,
        primitive_lattice_vectors,
        (permittivity,),
    ) = _validate_and_broadcast(
        wavelength, in_plane_wavevector, primitive_lattice_vectors, permittivity
    )
    (
        inverse_z_permittivity_matrix,
        z_permittivity_matrix,
        transverse_permittivity_matrix,
        tangent_vector_field,
    ) = fourier_matrices_patterned_isotropic_media(
        primitive_lattice_vectors=primitive_lattice_vectors,
        permittivity=permittivity,
        expansion=expansion,
        formulation=formulation,
    )

    # Create permeability matrices for nonmagnetic materials.
    ones = jnp.ones(
        z_permittivity_matrix.shape[:-1],
        dtype=z_permittivity_matrix.dtype,
    )
    zeros = jnp.zeros_like(ones)
    z_permeability_matrix = utils.diag(ones)
    inverse_z_permeability_matrix = utils.diag(ones)
    transverse_permeability_matrix = jnp.block(
        [
            [utils.diag(ones), utils.diag(zeros)],
            [utils.diag(zeros), utils.diag(ones)],
        ]
    )

    return _numerical_eigensolve(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        z_permittivity_matrix=z_permittivity_matrix,
        inverse_z_permittivity_matrix=inverse_z_permittivity_matrix,
        transverse_permittivity_matrix=transverse_permittivity_matrix,
        z_permeability_matrix=z_permeability_matrix,
        inverse_z_permeability_matrix=inverse_z_permeability_matrix,
        transverse_permeability_matrix=transverse_permeability_matrix,
        expansion=expansion,
        tangent_vector_field=tangent_vector_field,
    )


def _eigensolve_uniform_general_anisotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivities: TensorComponents,
    permeabilities: TensorComponents,
    expansion: basis.Expansion,
) -> LayerSolveResult:
    """Returns the results of a uniform anisotropic layer eigensolve.

    The layer is uniform and anisotropic, in the sense that the permittivity does not
    vary spatially and has orientation dependence.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivities: The elements of the permittivity tensor: `(eps_xx, eps_xy,
            eps_yx, eps_yy, eps_zz)`, each having shape `(..., nx, ny)`.
        permeabilities: The elements of the permeability tensor: `(mu_xx, mu_xy,
            mu_yx, mu_yy, mu_zz)`, each having shape `(..., nx, ny)`.
        expansion: The field expansion to be used.

    Returns:
        The `LayerSolveResult`.
    """
    if not all([p.shape[-2:] == (1, 1) for p in permittivities + permeabilities]):
        raise ValueError(
            f"Trailing axes of arrays in `permittivities` and `permeabilities` must "
            f"have shape (1, 1) but got a shapes {[p.shape for p in permittivities]} "
            f"and {[p.shape for p in permeabilities]}."
        )
    (
        wavelength,
        in_plane_wavevector,
        primitive_lattice_vectors,
        (
            permittivity_xx,
            permittivity_xy,
            permittivity_yx,
            permittivity_yy,
            permittivity_zz,
            permeability_xx,
            permeability_xy,
            permeability_yx,
            permeability_yy,
            permeability_zz,
        ),
    ) = _validate_and_broadcast(
        wavelength,
        in_plane_wavevector,
        primitive_lattice_vectors,
        *permittivities,
        *permeabilities,
    )

    batch_shape = jnp.broadcast_shapes(
        wavelength.shape, in_plane_wavevector.shape[:-1], permittivity_xx.shape[:-2]
    )
    shape = batch_shape + (expansion.num_terms,)

    permittivity_xx = jnp.broadcast_to(jnp.squeeze(permittivity_xx, axis=-1), shape)
    permittivity_xy = jnp.broadcast_to(jnp.squeeze(permittivity_xy, axis=-1), shape)
    permittivity_yx = jnp.broadcast_to(jnp.squeeze(permittivity_yx, axis=-1), shape)
    permittivity_yy = jnp.broadcast_to(jnp.squeeze(permittivity_yy, axis=-1), shape)
    permittivity_zz = jnp.broadcast_to(jnp.squeeze(permittivity_zz, axis=-1), shape)
    z_permittivity_matrix = utils.diag(permittivity_zz)
    inverse_z_permittivity_matrix = utils.diag(1 / permittivity_zz)
    # Note that the matrix element ordering and signs differ from [2012 Liu]
    # equation 37, but are consistent with the definition in equation 15. Equation 37
    # is likely in error.
    transverse_permittivity_matrix = jnp.block(
        [
            [utils.diag(permittivity_yy), utils.diag(-permittivity_yx)],
            [utils.diag(-permittivity_xy), utils.diag(permittivity_xx)],
        ]
    )

    permeability_xx = jnp.broadcast_to(jnp.squeeze(permeability_xx, axis=-1), shape)
    permeability_xy = jnp.broadcast_to(jnp.squeeze(permeability_xy, axis=-1), shape)
    permeability_yx = jnp.broadcast_to(jnp.squeeze(permeability_yx, axis=-1), shape)
    permeability_yy = jnp.broadcast_to(jnp.squeeze(permeability_yy, axis=-1), shape)
    permeability_zz = jnp.broadcast_to(jnp.squeeze(permeability_zz, axis=-1), shape)
    z_permeability_matrix = utils.diag(permeability_zz)
    inverse_z_permeability_matrix = utils.diag(1 / permeability_zz)
    # Note that the matrix element ordering for the transverse permittivity and
    # permeability matrices differs.
    transverse_permeability_matrix = jnp.block(
        [
            [utils.diag(permeability_xx), utils.diag(permeability_xy)],
            [utils.diag(permeability_yx), utils.diag(permeability_yy)],
        ]
    )

    return _numerical_eigensolve(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        z_permittivity_matrix=z_permittivity_matrix,
        inverse_z_permittivity_matrix=inverse_z_permittivity_matrix,
        transverse_permittivity_matrix=transverse_permittivity_matrix,
        z_permeability_matrix=z_permeability_matrix,
        inverse_z_permeability_matrix=inverse_z_permeability_matrix,
        transverse_permeability_matrix=transverse_permeability_matrix,
        expansion=expansion,
        tangent_vector_field=None,
    )


def _eigensolve_patterned_general_anisotropic_media(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivities: TensorComponents,
    permeabilities: TensorComponents,
    expansion: basis.Expansion,
    formulation: Formulation | VectorFn,
    vector_field_source: jnp.ndarray,
) -> LayerSolveResult:
    """Returns the results of a patterned anisotropic layer eigensolve.

    The layer is patterned and anisotropic, in the sense that the permittivity varies
    spatially and has orientation dependence.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivities: The elements of the permittivity tensor: `(eps_xx, eps_xy,
            eps_yx, eps_yy, eps_zz)`, each having shape `(..., nx, ny)`.
        permeabilities: The elements of the permeability tensor: `(mu_xx, mu_xy,
            mu_yx, mu_yy, mu_zz)`, each having shape `(..., nx, ny)`.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used, or a callable which computes
            the tangent vector field for a custom vector FMM formulation.
        vector_field_source: Array used to calculate the vector field, with shape
            matching the permittivities and permeabilities.

    Returns:
        The `LayerSolveResult`.
    """
    (
        wavelength,
        in_plane_wavevector,
        primitive_lattice_vectors,
        (
            permittivity_xx,
            permittivity_xy,
            permittivity_yx,
            permittivity_yy,
            permittivity_zz,
            permeability_xx,
            permeability_xy,
            permeability_yx,
            permeability_yy,
            permeability_zz,
            vector_field_source,
        ),
    ) = _validate_and_broadcast(
        wavelength,
        in_plane_wavevector,
        primitive_lattice_vectors,
        *permittivities,
        *permeabilities,
        vector_field_source,
    )
    (
        inverse_z_permittivity_matrix,
        z_permittivity_matrix,
        transverse_permittivity_matrix,
        inverse_z_permeability_matrix,
        z_permeability_matrix,
        transverse_permeability_matrix,
        tangent_vector_field,
    ) = fourier_matrices_patterned_anisotropic_media(
        primitive_lattice_vectors=primitive_lattice_vectors,
        permittivities=(
            permittivity_xx,
            permittivity_xy,
            permittivity_yx,
            permittivity_yy,
            permittivity_zz,
        ),
        permeabilities=(
            permeability_xx,
            permeability_xy,
            permeability_yx,
            permeability_yy,
            permeability_zz,
        ),
        expansion=expansion,
        formulation=formulation,
        vector_field_source=vector_field_source,
    )
    return _numerical_eigensolve(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        z_permittivity_matrix=z_permittivity_matrix,
        inverse_z_permittivity_matrix=inverse_z_permittivity_matrix,
        transverse_permittivity_matrix=transverse_permittivity_matrix,
        z_permeability_matrix=z_permeability_matrix,
        inverse_z_permeability_matrix=inverse_z_permeability_matrix,
        transverse_permeability_matrix=transverse_permeability_matrix,
        expansion=expansion,
        tangent_vector_field=tangent_vector_field,
    )


# -----------------------------------------------------------------------------
# Helper function used by all eigensolves done numerically.
# -----------------------------------------------------------------------------


def _numerical_eigensolve(
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    z_permittivity_matrix: jnp.ndarray,
    inverse_z_permittivity_matrix: jnp.ndarray,
    transverse_permittivity_matrix: jnp.ndarray,
    z_permeability_matrix: jnp.ndarray,
    inverse_z_permeability_matrix: jnp.ndarray,
    transverse_permeability_matrix: jnp.ndarray,
    expansion: basis.Expansion,
    tangent_vector_field: Optional[Tuple[jnp.ndarray, jnp.ndarray]],
) -> LayerSolveResult:
    r"""Returns the results of a patterned layer eigensolve.

    The layer may be anisotropic and magnetic, as determined by the provided transverse
    permittivity and permeability matrices.

    Args:
        wavelength: The free space wavelength of the excitation.
        in_plane_wavevector: `(kx0, ky0)`.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        inverse_z_permittivity_matrix: The fourier-transformed inverse of zz-component
            of permittivity.
        z_permittivity_matrix: The fourier-transformed zz-component of permittivity.
        transverse_permittivity_matrix: The fourier-transformed transverse permittivity
            matrix from equation 15 of [2012 Liu].
        inverse_z_permeability_matrix: The fourier-transformed inverse of zz-component
            of permeability.
        z_permeability_matrix: The fourier-transformed zz-component of permeability.
        transverse_permeability_matrix: The fourier-transformed transverse permeability
            matrix.
        expansion: The field expansion to be used.
        tangent_vector_field: The tangent vector field `(tx, ty)` used to compute the
            transverse permittivity matrix, if a vector FMM formulation is used. If
            the `FFT` formulation is used, the vector field is `None`.

    Returns:
        The `LayerSolveResult`.
    """
    # Transverse wavevectors are the `k + G` from equation 5 of [2012 Liu].
    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )

    # The k matrix from equation 23 of [2012 Liu], modified for magnetic materials.
    k_matrix = fmm_matrices.k_matrix_patterned(
        z_permeability_matrix, transverse_wavevectors
    )

    omega_script_k_matrix = fmm_matrices.omega_script_k_matrix_patterned(
        wavelength=wavelength,
        z_permittivity_matrix=z_permittivity_matrix,
        transverse_permeability_matrix=transverse_permeability_matrix,
        transverse_wavevectors=transverse_wavevectors,
    )

    # The matrix from equation 28 of [2012 Liu], modified for magnetic materials.
    matrix = (
        transverse_permittivity_matrix @ omega_script_k_matrix
        - k_matrix @ transverse_permeability_matrix
    )
    eigenvalues_squared, eigenvectors = utils.eig(matrix)
    eigenvalues = jnp.sqrt(eigenvalues_squared)
    eigenvalues = _select_eigenvalues_sign(eigenvalues)
    return LayerSolveResult(
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        eigenvalues=eigenvalues,
        eigenvectors=eigenvectors,
        z_permittivity_matrix=z_permittivity_matrix,
        inverse_z_permittivity_matrix=inverse_z_permittivity_matrix,
        z_permeability_matrix=z_permeability_matrix,
        inverse_z_permeability_matrix=inverse_z_permeability_matrix,
        transverse_permeability_matrix=transverse_permeability_matrix,
        tangent_vector_field=tangent_vector_field,
    )


# -----------------------------------------------------------------------------
# Functions for computing Fourier convolution matrices .
# -----------------------------------------------------------------------------


def fourier_matrices_patterned_isotropic_media(
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: Formulation | VectorFn,
) -> Tuple[
    jnp.ndarray, jnp.ndarray, jnp.ndarray, Optional[Tuple[jnp.ndarray, jnp.ndarray]]
]:
    """Return Fourier convolution matrices for patterned nonmagnetic isotropic media.

    All matrices are forms of the Fourier convolution matrices defined in equation
    8 of [2012 Liu]. For vector formulations, the transverse permittivity matrix is
    of the form E2 given in equation 51 of [2012 Liu].

    Args:
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivity: The permittivity array, with shape `(..., nx, ny)`.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used, or a callable which computes
            the tangent vector field for a custom vector FMM formulation.

    Returns:
        inverse_z_permittivity_matrix: The Fourier convolution matrix for the inverse
            of the z-component of the permittivity.
        z_permittivity_matrix: The Fourier convolution matrix for the z-component
            of the permittivity.
        transverse_permittivity_matrix: The transverse permittivity matrix.
        tangent_vector_field: The tangent vector field `(tx, ty)` used to compute the
            transverse permittivity matrix, if a vector FMM formulation is used. If
            the `FFT` formulation is used, the vector field is `None`.
    """
    if formulation is Formulation.FFT:
        _transverse_permittivity_fn = functools.partial(
            fmm_matrices.transverse_permittivity_fft,
            expansion=expansion,
        )
        tangent_vector_field = None
    else:
        if isinstance(formulation, Formulation):
            vector_fn = vector.VECTOR_FIELD_SCHEMES[formulation.value]
        else:
            vector_fn = formulation
        tx, ty = vector_fn(permittivity, expansion, primitive_lattice_vectors)
        _transverse_permittivity_fn = functools.partial(
            fmm_matrices.transverse_permittivity_vector,
            tx=tx,
            ty=ty,
            expansion=expansion,
        )
        tangent_vector_field = (tx, ty)

    _transform = functools.partial(fft.fourier_convolution_matrix, expansion=expansion)

    inverse_z_permittivity_matrix = _transform(1 / permittivity)
    z_permittivity_matrix = _transform(permittivity)
    transverse_permittivity_matrix = _transverse_permittivity_fn(permittivity)

    return (
        inverse_z_permittivity_matrix,
        z_permittivity_matrix,
        transverse_permittivity_matrix,
        tangent_vector_field,
    )


def fourier_matrices_patterned_anisotropic_media(
    primitive_lattice_vectors: basis.LatticeVectors,
    permittivities: TensorComponents,
    permeabilities: TensorComponents,
    expansion: basis.Expansion,
    formulation: Formulation | VectorFn,
    vector_field_source: jnp.ndarray,
) -> Tuple[
    jnp.ndarray,
    jnp.ndarray,
    jnp.ndarray,
    jnp.ndarray,
    jnp.ndarray,
    jnp.ndarray,
    Optional[Tuple[jnp.ndarray, jnp.ndarray]],
]:
    """Return Fourier convolution matrices for patterned anisotropic media.

    The transverse permittivity matrix E is defined as,

        [-Dy, Dx]^T = E [-Ey, Ex]^T

    while the transverse permeability matrix M is defined as,

        [Bx, By]^T = M [Hx, Hy]^T

    The Fourier factorization is done as for E1 given in equation 47 of [2012 Liu].

    Args:
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        permittivities: The elements of the permittivity tensor: `(eps_xx, eps_xy,
            eps_yx, eps_yy, eps_zz)`, each having shape `(..., nx, ny)`.
        permeabilities: The elements of the permeability tensor: `(mu_xx, mu_xy,
            mu_yx, mu_yy, mu_zz)`, each having shape `(..., nx, ny)`.
        expansion: The field expansion to be used.
        formulation: Specifies the formulation to be used.
        vector_field_source: Array used to calculate the vector field, with shape
            matching the permittivities and permeabilities.

    Returns:
        inverse_z_permittivity_matrix: The Fourier convolution matrix for the inverse
            of the z-component of the permittivity.
        z_permittivity_matrix: The Fourier convolution matrix for the z-component
            of the permittivity.
        transverse_permittivity_matrix: The transverse permittivity matrix from
            equation 15 of [2012 Liu], computed in the manner prescribed by
            `fmm_formulation`.
        inverse_z_permeability_matrix: The Fourier convolution matrix for the inverse
            of the z-component of the permeability.
        z_permeability_matrix: The Fourier convolution matrix for the z-component
            of the permeability.
        transverse_permeability_matrix: The transverse permittivity matrix.
        tangent_vector_field: The tangent vector field `(tx, ty)` used to compute the
            transverse permittivity matrix, if a vector FMM formulation is used. If
            the `FFT` formulation is used, the vector field is `None`.
    """
    if formulation is Formulation.FFT:
        _transverse_permittivity_fn = functools.partial(
            fmm_matrices.transverse_permittivity_fft_anisotropic,
            expansion=expansion,
        )
        _transverse_permeability_fn = functools.partial(
            fmm_matrices.transverse_permeability_fft_anisotropic,
            expansion=expansion,
        )
        tangent_vector_field = None
    else:
        if isinstance(formulation, Formulation):
            vector_fn = vector.VECTOR_FIELD_SCHEMES[formulation.value]
        else:
            vector_fn = formulation
        tx, ty = vector_fn(vector_field_source, expansion, primitive_lattice_vectors)
        _transverse_permittivity_fn = functools.partial(
            fmm_matrices.transverse_permittivity_vector_anisotropic,
            tx=tx,
            ty=ty,
            expansion=expansion,
        )
        _transverse_permeability_fn = functools.partial(
            fmm_matrices.transverse_permeability_vector_anisotropic,
            tx=tx,
            ty=ty,
            expansion=expansion,
        )
        tangent_vector_field = (tx, ty)

    _transform = functools.partial(fft.fourier_convolution_matrix, expansion=expansion)

    (
        permittivity_xx,
        permittivity_xy,
        permittivity_yx,
        permittivity_yy,
        permittivity_zz,
    ) = permittivities
    inverse_z_permittivity_matrix = _transform(1 / permittivity_zz)
    z_permittivity_matrix = _transform(permittivity_zz)
    transverse_permittivity_matrix = _transverse_permittivity_fn(
        permittivity_xx=permittivity_xx,
        permittivity_xy=permittivity_xy,
        permittivity_yx=permittivity_yx,
        permittivity_yy=permittivity_yy,
    )

    (
        permeability_xx,
        permeability_xy,
        permeability_yx,
        permeability_yy,
        permeability_zz,
    ) = permeabilities
    inverse_z_permeability_matrix = _transform(1 / permeability_zz)
    z_permeability_matrix = _transform(permeability_zz)
    transverse_permeability_matrix = _transverse_permeability_fn(
        permeability_xx=permeability_xx,
        permeability_xy=permeability_xy,
        permeability_yx=permeability_yx,
        permeability_yy=permeability_yy,
    )

    return (
        inverse_z_permittivity_matrix,
        z_permittivity_matrix,
        transverse_permittivity_matrix,
        inverse_z_permeability_matrix,
        z_permeability_matrix,
        transverse_permeability_matrix,
        tangent_vector_field,
    )


# -----------------------------------------------------------------------------
# Helper functions for validation and matrix assembly.
# -----------------------------------------------------------------------------


def _validate_and_broadcast(
    angular_frequency: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    *permittivities: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, basis.LatticeVectors, Tuple[jnp.ndarray, ...]]:
    """Validates that shapes are compatible and adds required batch dimensions."""
    if not in_plane_wavevector.shape[-1] == 2:
        raise ValueError(
            f"`in_plane_wavevector` must have a final dimension of size 2 but got "
            f"a shape of {in_plane_wavevector.shape}."
        )

    if not all([permittivities[0].shape == p.shape for p in permittivities]):
        raise ValueError("Got permittivities with differing shapes.")

    permittivity = permittivities[0]
    if not utils.batch_compatible_shapes(
        angular_frequency.shape,
        in_plane_wavevector.shape[:-1],
        primitive_lattice_vectors.u.shape[:-1],
        primitive_lattice_vectors.v.shape[:-1],
        permittivity.shape[:-2],
    ):
        raise ValueError(
            f"`angular_frequency`, `in_plane_wavevector`, `primitive_lattice_vectors` "
            f"and `permittivity` must be batch-compatible, but got shapes of "
            f"{angular_frequency.shape}, {in_plane_wavevector.shape}, "
            f"{primitive_lattice_vectors.u.shape}, "
            f"{primitive_lattice_vectors.v.shape}, and {permittivity.shape}."
        )

    num_batch_dims = max(
        [
            angular_frequency.ndim,
            in_plane_wavevector.ndim - 1,
            primitive_lattice_vectors.u.ndim - 1,
            primitive_lattice_vectors.v.ndim - 1,
            permittivity.ndim - 2,
        ]
    )
    angular_frequency = utils.atleast_nd(angular_frequency, n=num_batch_dims)
    in_plane_wavevector = utils.atleast_nd(in_plane_wavevector, n=num_batch_dims + 1)
    primitive_lattice_vectors = basis.LatticeVectors(
        u=utils.atleast_nd(primitive_lattice_vectors.u, n=num_batch_dims + 1),
        v=utils.atleast_nd(primitive_lattice_vectors.v, n=num_batch_dims + 1),
    )

    permittivities = tuple(
        [utils.atleast_nd(p, n=num_batch_dims + 2) for p in permittivities]
    )
    return (
        angular_frequency,
        in_plane_wavevector,
        primitive_lattice_vectors,
        permittivities,
    )


def _select_eigenvalues_sign(eigenvalues: jnp.ndarray) -> jnp.ndarray:
    """Selects the sign of eigenvalues to have strictly positive imaginary part.

    Args:
        eigenvalues: The eigenvalues whose sign is to be adjusted.

    Returns:
        The eigenvalues with adjusted sign.
    """
    return jnp.where(jnp.imag(eigenvalues) < 0, -eigenvalues, eigenvalues)


# -----------------------------------------------------------------------------
# Register custom objects in this module with jax to enable `jit`.
# -----------------------------------------------------------------------------


tree_util.register_pytree_node(
    Formulation,
    lambda x: ((), x.value),
    lambda value, _: Formulation(value),
)


tree_util.register_pytree_node(
    LayerSolveResult,
    lambda x: (
        (
            x.wavelength,
            x.in_plane_wavevector,
            x.primitive_lattice_vectors,
            x.expansion,
            x.eigenvalues,
            x.eigenvectors,
            x.z_permittivity_matrix,
            x.inverse_z_permittivity_matrix,
            x.z_permeability_matrix,
            x.inverse_z_permeability_matrix,
            x.transverse_permeability_matrix,
            x.tangent_vector_field,
        ),
        None,
    ),
    lambda _, x: LayerSolveResult(*x),
)
